@@ -31,6 +31,7 @@ import sys
 import re
 import traceback
 import glob
+import numpy as np
 
 from casacore.tables.tableutil import tableexists
 import Owlcat.Tables
@@ -141,6 +142,22 @@ def parse_subset_options(options):
         subset[opt] = getattr(options, opt)
     return subset
 
+def copy_chunked_flags(tab_from, tab_to, chunk_size):
+    """copies FLAG column in a chunked way. Returns two sums"""
+    nrows = tab_from.nrows()
+    fc_sum = fc_size = 0
+    for row0 in range(0, nrows, chunk_size):
+        n = min(nrows - row0, chunk_size)
+        fr = tab_from.getcol("FLAG_ROW", row0, n)
+        fc = tab_from.getcol("FLAG", row0, n)
+        fc[fr,:,:] = True
+        if tab_to is not None:
+            tab_to.putcol("FLAG_ROW", fr, row0, n)
+            tab_to.putcol("FLAG", fc, row0, n)
+        fc_sum += fc.sum(dtype=np.int64)
+        fc_size += fc.size
+    return fc_sum / float(fc_size)
+
 
 if __name__ == "__main__":
 
@@ -249,7 +266,11 @@ if __name__ == "__main__":
     group.add_option("--import", type="string", dest="_import", metavar="FILENAME",
                      help="imports flags from flag file. If any flagging actions are specified, these will be done after the import.")
     group.add_option("-R", "--restore", type="string", dest="restore", metavar="VERSION",
-                     help="imports flags from a CASA-style flagversion (use --list to list version names).")
+                     help="restores flags from a CASA-style flagversion (use --list to list version names).")
+    group.add_option("--save", type="string", dest="save", metavar="VERSION",
+                     help="saves flags to a CASA-style flagversion.")
+    group.add_option("--force", action="store_true",
+                     help="with --save: force overwrite of a CASA-style flagversion, if it exists.")
     group.add_option("-v", "--verbose", metavar="LEVEL", type="int",
                      help="verbosity level for messages. Higher is more verbose, default is 0.")
     group.add_option("--timestamps", action="store_true",
@@ -267,7 +288,7 @@ if __name__ == "__main__":
     (options, args) = parser.parse_args()
     if len(args) != 1:
         parser.error("Incorrect number of arguments. Use '-h' for help.")
-    msname = args[0]
+    msname = args[0].rstrip("/")
 
     import Owlcat
 
@@ -295,14 +316,9 @@ if __name__ == "__main__":
         try:
             ms = get_ms(readonly=False)
             ftab = Owlcat.table(flagvers)
-            flag_row = ftab.getcol("FLAG_ROW")
-            flag_col = ftab.getcol("FLAG")
+            ratio = copy_chunked_flags(ftab, ms, options.chunk_size)
             ftab.close()
-            ms.putcol("FLAG_ROW", flag_row)
-            ms.putcol("FLAG", flag_col)
-            flag_col[flag_row,...] = True
             ms.close()
-            ratio = flag_col.sum() / float(flag_col.size)
             print(f"Flag version '{options.restore}' restored: {ratio:.2%} flagged.")
         except:
             traceback.print_exc()
@@ -385,7 +401,7 @@ if __name__ == "__main__":
             # now check for CASA flagversions
             flagvers = [tab for tab in glob.glob(f"{msname}.flagversions/flags.*") if Owlcat.tableexists(tab)]
             if flagvers:
-                print(f"  {len(flagvers)} CASA-style flagversions available for --restore")
+                print(f"  {len(flagvers)} CASA-style flagversions available for --restore:")
                 for vers in flagvers:
                     name = os.path.basename(vers).split('.', 1)[1]
                     print(f"    '{name}'")
@@ -531,6 +547,22 @@ if __name__ == "__main__":
                         nvis_A, nvis_A * percent))
                 print(
                     "===>   %-29s includes %10d visibilities (%.3g%% of selection)" % (label, nv, nv * percent))
+                    
+            # now check for stats over CASA flagversions
+            flagvers = [tab for tab in glob.glob(f"{msname}.flagversions/flags.*") if Owlcat.tableexists(tab)]
+            if flagvers:
+                print(f"Stats for CASA-style flagversions:")
+                for vers in flagvers:
+                    name = os.path.basename(vers).split('.', 1)[1]
+                    ftab = Owlcat.table(vers, ack=False)
+                    ratio = copy_chunked_flags(ftab, None, options.chunk_size)
+                    ftab.close()
+                    print(f"    '{name}': {ratio:.2%} of all data flagged.")
+            else:
+                print("  No CASA-style flagversions available.")
+            if options.flag or options.unflag or options.copy or options.fill_legacy or options.remove:
+                print("-l/--list was in effect, so all other options were ignored.")
+            print("")
             sys.exit(0)
 
         # else not stats mode, do the actual flagging job
@@ -585,3 +617,44 @@ if __name__ == "__main__":
             traceback.print_exc()
             error("Error exporting flags to %s" % options.export)
         print("Flags exported OK.")
+
+    if options.save:
+        flagvers = f"{msname}.flagversions/flags.{options.save}"
+        if Owlcat.tableexists(flagvers) and not options.force:
+            error(f"Flagversion '{options.save}' already exists. Please run with --force to confirm overwrite.")
+        try:
+            ms = get_ms(readonly=True)
+            fr_desc = ms.getcoldesc("FLAG_ROW")
+            fc_desc = ms.getcoldesc("FLAG")
+            if os.path.exists(flagvers):
+                os.system(f"rm -fr {flagvers}")
+            from casacore.tables import maketabdesc
+            fr_desc['name'] = "FLAG_ROW"
+            fc_desc['name'] = "FLAG"
+            td = maketabdesc([dict(name='FLAG_ROW', desc=fr_desc), dict(name='FLAG', desc=fc_desc)])
+            ftab = Owlcat.table(flagvers, tabledesc=td, readonly=False, nrow=ms.nrows())
+            ratio = copy_chunked_flags(ms, ftab, options.chunk_size)
+            ftab.close()
+            ms.close()
+            print(f"Flag version '{options.save}' saved: {ratio:.2%} flagged.")
+            # add to list
+            try:
+                version_list_file = f"{msname}.flagversions/FLAG_VERSION_LIST"
+                version_list = open(version_list_file, encoding="ascii").readlines()
+                versions = set(line.split(':', 1)[0].strip() for line in  version_list)
+                if options.save in versions:
+                    print(f"Flag version '{options.save}' is present in CASA's FLAG_VERSION_LIST.")
+                else:
+                    with open(version_list_file, "wt", encoding="ascii") as fo:
+                        for line in version_list:
+                            fo.write(line.rstrip() + "\n")
+                        fo.write(f"{options.save} :\n")
+                    print(f"Flag version '{options.save}' added to CASA's FLAG_VERSION_LIST.")
+            except Exception:
+                traceback.print_exc()
+                print("There was an error writing CASA's FLAG_VERSION_LIST file.")
+                print("Your new flag version might not be recognized by CASA's flagmanager.")
+
+        except:
+            traceback.print_exc()
+            error(f"Error saving flags to {options.save}, exiting")
